/*
 * Copyright (C) 2020 Linux Studio Plugins Project <https://lsp-plug.in/>
 *           (C) 2020 Vladimir Sadovnikov <sadko4u@gmail.com>
 *
 * This file is part of lsp-plugins
 * Created on: 21 янв. 2019 г.
 *
 * lsp-plugins is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 *
 * lsp-plugins is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with lsp-plugins. If not, see <https://www.gnu.org/licenses/>.
 */

#include <core/3d/common.h>

namespace lsp
{
#ifdef LSP_DEBUG
    const color3d_t C3D_RED         = { 1.0f, 0.0f, 0.0f, 1.0f };
    const color3d_t C3D_GREEN       = { 0.0f, 1.0f, 0.0f, 1.0f };
    const color3d_t C3D_DARKGREEN   = { 0.0f, 0.75f, 0.0f, 1.0f };
    const color3d_t C3D_BLUE        = { 0.0f, 0.0f, 1.0f, 1.0f };
    const color3d_t C3D_CYAN        = { 0.0f, 1.0f, 1.0f, 1.0f };
    const color3d_t C3D_MAGENTA     = { 1.0f, 0.0f, 1.0f, 1.0f };
    const color3d_t C3D_YELLOW      = { 1.0f, 1.0f, 0.0f, 1.0f };
    const color3d_t C3D_ORANGE      = { 1.0f, 0.5f, 0.0f, 1.0f };
    const color3d_t C3D_GRAY        = { 0.75f, 0.75f, 0.75f, 1.0f };
#endif /* LSP_DEBUG */
}


