<p><img src="<?= $DOCROOT ?>img/logo.png" style="width:20%; height:20%"></p>
<p><b>LSP (Linux Studio Plugins)</b> is a collection of open-source plugins compatible with following formats:</p>
<ul>
	<li><b>LADSPA</b> - set of plugins for <a href="https://en.wikipedia.org/wiki/LADSPA">Linux Audio Developer's Simple Plugin API</a></li>
	<li><b>LV2</b> - set of plugins and UIs for <a href="http://lv2plug.in/">Linux Audio Developer's Simple Plugin API (LADSPA) version 2</a></li>
	<li><b>LinuxVST</b> - set of plugins and UIs for <a href="http://www.steinberg.net/">Steinberg's VST 2.4</a> format ported on GNU/Linux Platform</li>
	<li><b>JACK</b> - Standalone versions for <a href="http://jackaudio.org/">JACK Audio connection Kit</a> with UI</li>
</ul>

<p>The basic idea is to fill the lack of good and useful plugins under the GNU/Linux platform.</p>

<p>After some contributions to other open source projects the decision was made to implement separate and independent plugin distribution.</p>

<p>Currently project is developed and maintained by LSP Project Team with the main founder and developer at the head &ndash; Vladimir Sadovnikov.</p>

<p>The official site ot the project is located here: <a href="http://lsp-plug.in/">http://lsp-plug.in/</a></p>
