<p>The LSP project is an open-source project and cares about quality of developed software.</p>

<p>Still there is no absolute warranty about stability of the software on different platforms, 
so you're using this software on your own risk.</p>

<p>The source code takes a lot of time to develop, so the project uses donation model to
compensate time spent for development.</p>

<p>Binary distributions of the software are completely free but the source code for each plugin
will be published under the free license only after the donation goal is reached. Detailed
information is available at official site: <a href="http://lsp-plug.in/">http://lsp-plug.in/</a></p>

<p>Below the text of license used by <b>LSP Plugins</b> is present.</p>

<pre>
<?php file_content('LICENSE.txt'); ?>
</pre>
